@@ -19,8 +19,6 @@ from __future__ import absolute_import
 from __future__ import print_function
 from __future__ import unicode_literals
 
-import sys
-
 from . import keywords
 from . import tokenize
 
@@ -56,9 +54,6 @@ FUNCTION_ATTRIBUTE = 0x20
 FUNCTION_UNKNOWN_ANNOTATION = 0x40
 FUNCTION_THROW = 0x80
 
-_INTERNAL_TOKEN = 'internal'
-_NAMESPACE_POP = 'ns-pop'
-
 
 class ParseError(Exception):
 
@@ -173,8 +168,9 @@ class _GenericDeclaration(Node):
 
     def full_name(self):
         prefix = ''
-        if self.namespace and self.namespace[-1]:
-            prefix = '::'.join(self.namespace) + '::'
+        names = [n for n in self.namespace if n is not None]
+        if names:
+            prefix = '::'.join(names) + '::'
         return prefix + self.name
 
     def _type_string_helper(self, suffix):
@@ -220,7 +216,7 @@ class Typedef(_GenericDeclaration):
         return self._type_string_helper(suffix)
 
 
-class _NestedType(_GenericDeclaration):
+class Enum(_GenericDeclaration):
 
     def __init__(self, start, end, name, fields, namespace):
         _GenericDeclaration.__init__(self, start, end, name, namespace)
@@ -237,14 +233,6 @@ class _NestedType(_GenericDeclaration):
         return self._type_string_helper(suffix)
 
 
-class Union(_NestedType):
-    pass
-
-
-class Enum(_NestedType):
-    pass
-
-
 class Class(_GenericDeclaration):
 
     def __init__(self, start, end, name,
@@ -276,6 +264,10 @@ class Struct(Class):
     pass
 
 
+class Union(Class):
+    pass
+
+
 class Function(_GenericDeclaration):
 
     def __init__(self, start, end, name, return_type, parameters,
@@ -383,7 +375,7 @@ class TypeConverter(object):
     def _get_template_end(self, tokens, start):
         count = 1
         end = start
-        while count:
+        while count and end < len(tokens):
             token = tokens[end]
             end += 1
             if token.name == '<':
@@ -468,9 +460,6 @@ class TypeConverter(object):
         default = []
         other_tokens = []
 
-        if parts[0].name == '::':
-            parts = parts[1:]
-
         # Handle default (initial) values properly.
         for i, t in enumerate(parts):
             if t.name == '[' and arrayBegin == 0:
@@ -520,9 +509,8 @@ class TypeConverter(object):
         if needs_name_removed:
             name = type_name.pop()
 
-        type_name = ' '.join([t for t in type_name])
         return (name,
-                type_name,
+                ' '.join([t for t in type_name]),
                 templated_types,
                 modifiers,
                 default,
@@ -549,12 +537,13 @@ class TypeConverter(object):
              _, __) = self.declaration_to_parts(type_modifiers,
                                                 True)
 
-            parameter_type = Type(first_token.start, first_token.end,
-                                  type_name, templated_types, modifiers,
-                                  reference, pointer, False)
-            p = Parameter(first_token.start, end, name,
-                          parameter_type, default)
-            result.append(p)
+            if type_name:
+                parameter_type = Type(first_token.start, first_token.end,
+                                      type_name, templated_types, modifiers,
+                                      reference, pointer, False)
+                p = Parameter(first_token.start, end, name,
+                              parameter_type, default)
+                result.append(p)
 
         template_count = 0
         for s in tokens:
@@ -633,23 +622,17 @@ class ASTBuilder(object):
         self.filename = filename
         self.token_queue = []
         self.namespace_stack = namespace_stack[:]
+        self.namespaces = []
         self.define = set()
         self.quiet = quiet
         self.in_class = in_class
-        self.in_class_name_only = None
-        if in_class is not None:
-            self.in_class_name_only = in_class.split('::')[-1].split('<')[0]
+        if in_class:
+            self.namespaces.append(False)
         # Keep the state whether we are currently handling a typedef or not.
         self._handling_typedef = False
         self._handling_const = False
         self.converter = TypeConverter(self.namespace_stack)
 
-    def handle_error(self, msg, token):
-        if not self.quiet:
-            printable_queue = list(reversed(self.token_queue[-20:]))
-            sys.stderr.write('Got %s in %s @ %s %s\n' %
-                             (msg, self.filename, token, printable_queue))
-
     def generate(self):
         while True:
             try:
@@ -657,12 +640,12 @@ class ASTBuilder(object):
             except StopIteration:
                 break
 
-            # Dispatch on the next token type.
-            if (
-                token.token_type == _INTERNAL_TOKEN and
-                token.name == _NAMESPACE_POP
-            ):
-                self.namespace_stack.pop()
+            if token.name == '{':
+                self.namespaces.append(False)
+                continue
+            if token.name == '}':
+                if self.namespaces.pop():
+                    self.namespace_stack.pop()
                 continue
 
             result = self._generate_one(token)
@@ -690,18 +673,6 @@ class ASTBuilder(object):
                 method = getattr(self, 'handle_' + token.name, None)
                 assert_parse(method, 'unexpected token: {}'.format(token))
                 return method()
-            elif token.name == self.in_class_name_only:
-                # The token name is the same as the class, must be a ctor if
-                # there is a paren. Otherwise, it's the return type.
-                # Peek ahead to get the next token to figure out which.
-                next_item = self._get_next_token()
-                self._add_back_token(next_item)
-                if (
-                    next_item.token_type == tokenize.SYNTAX and
-                    next_item.name == '('
-                ):
-                    return self._get_method([token], FUNCTION_CTOR, None, True)
-                # Fall through--handle like any other method.
 
             # Handle data or function declaration/definition.
             temp_tokens, last_token = \
@@ -716,6 +687,11 @@ class ASTBuilder(object):
                     self._ignore_up_to(';')
                     return None
 
+                # Ignore __declspec
+                if temp_tokens[-1].name == '__declspec':
+                    self._ignore_up_to(')')
+                    return None
+
                 # If there is an assignment before the paren,
                 # this is an expression, not a method.
                 for i, elt in reversed(list(enumerate(temp_tokens))):
@@ -730,20 +706,7 @@ class ASTBuilder(object):
                         break
 
             if last_token.name == ';':
-                # Handle data, this isn't a method.
-                name, type_name, templated_types, modifiers, default, _ = \
-                    self.converter.declaration_to_parts(temp_tokens, True)
-
-                assert_parse(temp_tokens, 'not enough tokens')
-
-                t0 = temp_tokens[0]
-                names = [t.name for t in temp_tokens]
-                if templated_types:
-                    start, end = self.converter.get_template_indices(names)
-                    names = names[:start] + names[end:]
-                default = ''.join([t.name for t in default])
-                return self._create_variable(t0, name, type_name, modifiers,
-                                             names, templated_types, default)
+                return self._get_variable(temp_tokens)
             if last_token.name == '{':
                 assert_parse(temp_tokens, 'not enough tokens')
 
@@ -759,11 +722,7 @@ class ASTBuilder(object):
             if token.name == '~' and self.in_class:
                 # Must be a dtor (probably not in method body).
                 token = self._get_next_token()
-                # self.in_class can contain A::Name, but the dtor will only
-                # be Name. Make sure to compare against the right value.
-                if (token.token_type == tokenize.NAME and
-                        token.name == self.in_class_name_only):
-                    return self._get_method([token], FUNCTION_DTOR, None, True)
+                return self._get_method([token], FUNCTION_DTOR, None, True)
             # TODO(nnorwitz): handle a lot more syntax.
         elif token.token_type == tokenize.PREPROCESSOR:
             # TODO(nnorwitz): handle more preprocessor directives.
@@ -834,14 +793,15 @@ class ASTBuilder(object):
                 count1 += 1
             elif last_token.name == ']':
                 count1 -= 1
-            if skip_bracket_content:
+            if skip_bracket_content and count1 == 0:
                 if last_token.name == 'operator':
                     skip_bracket_content = False
                 elif last_token.name == '<':
                     count2 += 1
                 elif last_token.name == '>':
                     count2 -= 1
-            tokens.append(last_token)
+            if last_token.token_type != tokenize.PREPROCESSOR:
+                tokens.append(last_token)
             temp_token = self._get_next_token()
             if temp_token.name == '(' and last_token.name in self.define:
                 # TODO: for now just ignore the tokens inside the parenthesis
@@ -859,18 +819,14 @@ class ASTBuilder(object):
         # Assumes the current token is open_paren and we will consume
         # and return up to the close_paren.
         count = 1
-        token = get_next_token()
-        while True:
+        while count != 0:
+            token = get_next_token()
             if token.token_type == tokenize.SYNTAX:
                 if token.name == open_paren:
                     count += 1
                 elif token.name == close_paren:
                     count -= 1
-                    if count == 0:
-                        break
             yield token
-            token = get_next_token()
-        yield token
 
     def _get_parameters(self):
         return self._get_matching_char('(', ')')
@@ -968,7 +924,7 @@ class ASTBuilder(object):
                     count -= 1
                 elif tok.name == '>':
                     count += 1
-            specializations = return_type_and_name[index+1:]
+            specializations = return_type_and_name[index + 1:]
             del return_type_and_name[index:]
             name = return_type_and_name.pop()
         elif name.name == ']':
@@ -986,6 +942,7 @@ class ASTBuilder(object):
         if name.name == self.in_class and not modifiers:
             modifiers |= FUNCTION_CTOR
         parameters = list(self._get_parameters())
+        assert_parse(parameters, 'missing closing parenthesis')
         last_token = parameters.pop()    # Remove trailing ')'.
 
         # Handling operator() is especially weird.
@@ -1016,7 +973,7 @@ class ASTBuilder(object):
                 modifiers |= FUNCTION_ATTRIBUTE
                 token = self._get_next_token()
                 assert_parse(token.name == '(', token)
-                # Consume everything between the (parens).
+                # Consume everything between the parens.
                 list(self._get_matching_char('(', ')'))
                 token = self._get_next_token()
             elif token.name == 'throw':
@@ -1054,42 +1011,38 @@ class ASTBuilder(object):
 
         # Handle ctor initializers.
         if token.name == ':':
-            # TODO(nnorwitz): anything else to handle for initializer list?
             while token.name != ';' and token.name != '{':
+                _, token = self.get_name()
+                if token.name == '(':
+                    list(self._get_matching_char('(', ')'))
+                elif token.name == '{':
+                    list(self._get_matching_char('{', '}'))
                 token = self._get_next_token()
 
         # Handle pointer to functions that are really data but look
         # like method declarations.
         if token.name == '(':
-            if parameters[0].name in '*&':
-                # name contains the return type.
-                return_type.append(name)
-                name = parameters.pop()
-                # Already at the ( to open the parameter list.
-                function_parameters = list(self._get_matching_char('(', ')'))
-                del function_parameters[-1]  # Remove trailing ')'.
-                # TODO(nnorwitz): store the function_parameters.
-                token = self._get_next_token()
+            # name contains the return type.
+            return_type.append(name)
+            name = parameters.pop()
+            # Already at the ( to open the parameter list.
+            function_parameters = list(self._get_matching_char('(', ')'))
+            del function_parameters[-1]  # Remove trailing ')'.
+            # TODO(nnorwitz): store the function_parameters.
+            token = self._get_next_token()
 
-                assert_parse(token.token_type == tokenize.SYNTAX, token)
-                assert_parse(token.name == ';', token)
+            default = []
+            if token.name == '=':
+                default.extend(self._get_tokens_up_to(';'))
 
-                types = [t.name for t in return_type]
-                return self._create_variable(indices, name.name, indices.name,
-                                             [], types)
-            # At this point, we got something like:
-            #  return_type (type::*name_)(params);
-            # This is a data member called name_ that is a function pointer.
-            # With this code: void (sq_type::*field_)(string&);
-            # We get: name=void return_type=[] parameters=sq_type ... field_
-            # TODO(nnorwitz): is return_type always empty?
-            # TODO(nnorwitz): this isn't even close to being correct.
-            # Just put in something so we don't crash and can move on.
-            real_name = parameters[-1]
-            modifiers = [p.name for p in self._get_parameters()]
-            del modifiers[-1]           # Remove trailing ')'.
-            return self._create_variable(indices, real_name.name, indices.name,
-                                         modifiers, '')
+            return self._create_variable(
+                indices,
+                name.name,
+                indices.name,
+                [],
+                [t.name for t in return_type],
+                None,
+                ''.join([t.name for t in default]))
 
         if token.name == '{':
             body = list(self.get_scope())
@@ -1108,6 +1061,14 @@ class ASTBuilder(object):
                 list(self._get_matching_char('[', ']'))
                 token = self._get_next_token()
 
+            if token.name in '*&':
+                tokens, last = self._get_var_tokens_up_to(False, '(', ';')
+                tokens.insert(0, token)
+                tokens = parameters + tokens
+                if last.name == '(':
+                    return self._get_method(tokens, 0, None, False)
+                return self._get_variable(tokens)
+
             assert_parse(token.name == ';',
                          (token, return_type_and_name, parameters))
 
@@ -1122,6 +1083,21 @@ class ASTBuilder(object):
                         parameters, specializations, modifiers,
                         templated_types, body, self.namespace_stack)
 
+    def _get_variable(self, tokens):
+        name, type_name, templated_types, modifiers, default, _ = \
+            self.converter.declaration_to_parts(tokens, True)
+
+        assert_parse(tokens, 'not enough tokens')
+
+        t0 = tokens[0]
+        names = [t.name for t in tokens]
+        if templated_types:
+            start, end = self.converter.get_template_indices(names)
+            names = names[:start] + names[end:]
+        default = ''.join([t.name for t in default])
+        return self._create_variable(t0, name, type_name, modifiers,
+                                     names, templated_types, default)
+
     def _get_return_type_and_class_name(self, token_seq):
         # Splitting the return type from the class name in a method
         # can be tricky. For example, Return::Type::Is::Hard::To::Find().
@@ -1157,7 +1133,37 @@ class ASTBuilder(object):
         class_name = names[-1]
         return return_type, class_name
 
-    def _get_nested_type(self, ctor):
+    def _handle_class_and_struct(self, class_type):
+        if self._handling_typedef:
+            return self._get_class(class_type, None)
+
+        name_tokens, var_token = self.get_name()
+        if var_token.token_type == tokenize.NAME or var_token.name in '*&':
+            tokens, last = self._get_var_tokens_up_to(False, '(', ';', '{')
+            tokens.insert(0, var_token)
+            tokens = name_tokens + tokens
+            if last.name == '{':
+                self._add_back_token(last)
+                self._add_back_tokens(tokens)
+                return self._get_class(class_type, None)
+            if last.name == '(':
+                return self._get_method(tokens, 0, None, False)
+            return self._get_variable(tokens)
+
+        self._add_back_token(var_token)
+        self._add_back_tokens(name_tokens)
+        return self._get_class(class_type, None)
+
+    def handle_class(self):
+        return self._handle_class_and_struct(Class)
+
+    def handle_struct(self):
+        return self._handle_class_and_struct(Struct)
+
+    def handle_union(self):
+        return self._handle_class_and_struct(Union)
+
+    def handle_enum(self):
         # Handle strongly typed enumerations.
         token = self._get_next_token()
         if token.name != 'class':
@@ -1171,18 +1177,15 @@ class ASTBuilder(object):
         if token.token_type == tokenize.NAME:
             if self._handling_typedef:
                 self._add_back_token(token)
-                return ctor(token.start, token.end, name, None,
+                return Enum(token.start, token.end, name, None,
                             self.namespace_stack)
 
             next_token = self._get_next_token()
             if next_token.name != '(':
                 self._add_back_token(next_token)
             else:
-                token = next_token
-
-        if token.token_type == tokenize.SYNTAX and token.name == '(':
-            self._ignore_up_to(')')
-            token = self._get_next_token()
+                name_tokens.append(token)
+                return self._get_method(name_tokens, 0, None, False)
 
         # Handle underlying type.
         if token.token_type == tokenize.SYNTAX and token.name == ':':
@@ -1190,7 +1193,7 @@ class ASTBuilder(object):
 
         # Handle forward declarations.
         if token.token_type == tokenize.SYNTAX and token.name == ';':
-            return ctor(token.start, token.end, name, None,
+            return Enum(token.start, token.end, name, None,
                         self.namespace_stack)
 
         # Must be the type declaration.
@@ -1198,7 +1201,7 @@ class ASTBuilder(object):
             fields = list(self._get_matching_char('{', '}'))
             del fields[-1]                  # Remove trailing '}'.
             next_item = self._get_next_token()
-            new_type = ctor(token.start, token.end, name, fields,
+            new_type = Enum(token.start, token.end, name, fields,
                             self.namespace_stack)
             # A name means this is an anonymous type and the name
             # is the variable declaration.
@@ -1211,64 +1214,6 @@ class ASTBuilder(object):
         assert_parse(token.token_type == tokenize.NAME, token)
         return self._create_variable(token, token.name, name, [], '')
 
-    def _handle_class_and_struct(self, class_type, class_str):
-        # Special case the handling typedef/aliasing of classes/structs here.
-        name_tokens, var_token = self.get_name()
-        if name_tokens and not self._handling_typedef:
-            # Forward declaration.
-            if var_token.name == ';':
-                return class_type(name_tokens[0].start, name_tokens[0].end,
-                                  name_tokens[0].name, None, None, None,
-                                  self.namespace_stack)
-
-            next_token = self._get_next_token()
-            is_syntax = (var_token.token_type == tokenize.SYNTAX and
-                         var_token.name[0] in '*&')
-            is_variable = (var_token.token_type == tokenize.NAME and
-                           next_token.name == ';')
-            variable = var_token
-            if is_syntax and not is_variable:
-                temp = next_token
-                while variable.token_type != tokenize.NAME:
-                    variable = temp
-                    temp = self._get_next_token()
-                if temp.token_type == tokenize.SYNTAX and temp.name == '(':
-                    # Handle methods declared to return a class/struct.
-                    t0 = name_tokens[0]
-                    token = tokenize.Token(tokenize.NAME, class_str,
-                                           t0.start - 7, t0.start - 2)
-                    type_and_name = [token]
-                    type_and_name.extend(name_tokens)
-                    type_and_name.extend((var_token, next_token))
-                    return self._get_method(type_and_name, 0, None, False)
-
-                assert_parse(temp.name == ';', (temp, name_tokens, var_token))
-
-            if is_syntax or is_variable:
-                modifiers = [class_str]
-                type_name = ''.join([t.name for t in name_tokens])
-                position = name_tokens[0]
-                return self._create_variable(
-                    position, variable.name, type_name,
-                    modifiers, var_token.name)
-            name_tokens.extend((var_token, next_token))
-        else:
-            self._add_back_token(var_token)
-        self._add_back_tokens(name_tokens)
-        return self._get_class(class_type, None)
-
-    def handle_class(self):
-        return self._handle_class_and_struct(Class, 'class')
-
-    def handle_struct(self):
-        return self._handle_class_and_struct(Struct, 'struct')
-
-    def handle_union(self):
-        return self._get_nested_type(Union)
-
-    def handle_enum(self):
-        return self._get_nested_type(Enum)
-
     def handle_const(self):
         self._handling_const = True
         token = self._get_next_token()
@@ -1309,7 +1254,7 @@ class ASTBuilder(object):
             tokens.append(last)
             self._add_back_tokens(tokens)
             token = self._get_next_token()
-            if token.name == '::':
+            while token.name in ('inline', 'typename', '::'):
                 token = self._get_next_token()
             result = self._generate_one(token)
         else:
@@ -1363,9 +1308,6 @@ class ASTBuilder(object):
         return Typedef(name.start, name.end, name.name,
                        new_type, self.namespace_stack)
 
-    def handle_typeid(self):
-        pass  # Not needed yet.
-
     def handle_typename(self):
         pass  # Not needed yet.
 
@@ -1447,12 +1389,12 @@ class ASTBuilder(object):
             bases_ast = self.converter.to_type(base)
             assert_parse(len(bases_ast) == 1, bases_ast)
             bases.append(bases_ast[0])
+            if next_token.token_type == tokenize.PREPROCESSOR:
+                next_token = self._get_next_token()
             assert_parse(next_token.token_type == tokenize.SYNTAX, next_token)
             if next_token.name == '{':
                 token = next_token
                 break
-            # Support multiple inheritance.
-            assert_parse(next_token.name == ',', next_token)
         return bases, token
 
     def _get_class(self, class_type, templated_types):
@@ -1477,8 +1419,10 @@ class ASTBuilder(object):
                 attribute, token = self.get_name()
                 if len(attribute) > 1 or attribute[0].name != 'final':
                     name_tokens = attribute
-            class_name = ''.join([t.name for t in name_tokens])
+            class_name = self.converter.to_type(name_tokens)[0].name
         bases = None
+        if token.token_type == tokenize.PREPROCESSOR:
+            token = self._get_next_token()
         if token.token_type == tokenize.SYNTAX:
             if token.name == ';':
                 # Forward declaration.
@@ -1517,6 +1461,7 @@ class ASTBuilder(object):
                     assert_parse(token.token_type == tokenize.SYNTAX, token)
                     assert_parse(token.name == ';', token)
                 else:
+                    self._ignore_up_to(';')
                     new_class = class_type(class_token.start, class_token.end,
                                            class_name, bases, None,
                                            body, self.namespace_stack)
@@ -1527,7 +1472,8 @@ class ASTBuilder(object):
                                                  modifiers, token.name)
         else:
             if not self._handling_typedef:
-                self.handle_error('non-typedef token', token)
+                name_tokens = [class_token] + name_tokens
+                return self._get_method(name_tokens, 0, None, False)
             self._add_back_token(token)
 
         return class_type(class_token.start, class_token.end, class_name,
@@ -1540,22 +1486,16 @@ class ASTBuilder(object):
         if token.token_type == tokenize.NAME:
             name = token.name
             token = self._get_next_token()
-        self.namespace_stack.append(name)
         assert_parse(token.token_type == tokenize.SYNTAX, token)
-        # Create an internal token that denotes when the namespace is complete.
-        internal_token = tokenize.Token(_INTERNAL_TOKEN, _NAMESPACE_POP,
-                                        None, None)
+
         if token.name == '=':
             # TODO(nnorwitz): handle aliasing namespaces.
             name, next_token = self.get_name()
             assert_parse(next_token.name == ';', next_token)
-            self._add_back_token(internal_token)
         else:
             assert_parse(token.name == '{', token)
-            tokens = list(self.get_scope())
-            tokens.append(internal_token)
-            # Handle namespace with nothing in it.
-            self._add_back_tokens(tokens)
+            self.namespace_stack.append(name)
+            self.namespaces.append(True)
         return None
 
     def handle_using(self):
