@@ -86,6 +86,9 @@ def is_cpp_file(filename):
 
 class WarningHunter(object):
 
+    # Cache filename: ast_list
+    _module_cache = {}
+
     def __init__(self, filename, source, ast_list, include_paths, quiet=False):
         self.filename = filename
         self.source = source
@@ -96,6 +99,8 @@ class WarningHunter(object):
 
         self.metrics = metrics.Metrics(source)
         self.warnings = set()
+        if filename not in self._module_cache:
+            self._module_cache[filename] = Module(filename, ast_list)
 
     def _add_warning(self, msg, node, filename=None):
         if filename is not None:
@@ -132,6 +137,12 @@ class WarningHunter(object):
             module = Module(filename, None)
             msg = "unable to find '{}'".format(filename)
             self._add_warning(msg, node)
+        elif filename in self._module_cache:
+            # The cache survives across all instances, but the symbol table
+            # is per instance, so we need to make sure the symbol table
+            # is updated even if the module was in the cache.
+            module = self._module_cache[filename]
+            self._update_symbol_table(module)
         else:
             ast_list = None
             try:
@@ -148,6 +159,7 @@ class WarningHunter(object):
                             error),
                         file=sys.stderr)
             module = Module(filename, ast_list)
+            self._module_cache[filename] = module
             self._update_symbol_table(module)
         return module
 
@@ -208,8 +220,6 @@ class WarningHunter(object):
         """Find all the forward declarations that are not used."""
         for cls in forward_declarations:
             if cls in file_uses:
-                # TODO: Avoid the false positive in the case where the forward
-                # declaration is of a class that is defined in that same file.
                 if not decl_uses[cls] & USES_DECLARATION:
                     node = forward_declarations[cls]
                     msg = ("'{}' forward declared, "
@@ -227,29 +237,40 @@ class WarningHunter(object):
         decl_uses = dict.fromkeys(forward_declarations, UNUSED)
         symbol_table = self.symbol_table
 
-        def _add_reference(name, namespace):
-            if name in decl_uses:
+        for name, node in forward_declarations.items():
+            try:
+                symbol_table.lookup_symbol(node.name, node.namespace)
                 decl_uses[name] |= USES_REFERENCE
-            else:
-                nss = ''
-                for ns in namespace:
-                    if ns is None:
-                        continue
-                    nss += ns + '::'
-                    if nss + name in decl_uses:
-                        decl_uses[nss + name] |= USES_REFERENCE
-                        return
+            except symbols.Error:
+                module = Module(name, None)
+                self.symbol_table.add_symbol(node.name, node.namespace, node,
+                                             module)
 
-                try:
-                    file_use_node = symbol_table.lookup_symbol(name, namespace)
-                except symbols.Error:
-                    return
-                name = file_use_node[1].filename
-                if name in file_uses:
-                    if isinstance(file_use_node[0], ast.Typedef):
-                        file_uses[name] |= USES_DECLARATION
-                    else:
-                        file_uses[name] |= USES_REFERENCE
+        def _add_declaration(name, namespace):
+            names = [n for n in namespace if n is not None]
+            if names:
+                name = '::'.join(names) + '::' + name
+            if name in decl_uses:
+                decl_uses[name] |= USES_DECLARATION
+
+        def _add_reference(name, namespace):
+            try:
+                file_use_node = symbol_table.lookup_symbol(name, namespace)
+            except symbols.Error:
+                return
+
+            name = file_use_node[1].filename
+            if file_use_node[1].ast_list is None:
+                decl_uses[name] |= USES_REFERENCE
+            elif name in file_uses:
+                # enum and typedef can't be forward declared
+                if (
+                    isinstance(file_use_node[0], ast.Enum) or
+                    isinstance(file_use_node[0], ast.Typedef)
+                ):
+                    file_uses[name] |= USES_DECLARATION
+                else:
+                    file_uses[name] |= USES_REFERENCE
 
         def _add_use(name, namespace):
             if isinstance(name, list):
@@ -262,20 +283,10 @@ class WarningHunter(object):
             try:
                 file_use_node = symbol_table.lookup_symbol(name, namespace)
             except symbols.Error:
-                # TODO(nnorwitz): symbols from the current module
-                # should be added to the symbol table and then this
-                # exception should not happen...unless the code relies
-                # on another header for proper compilation.
-                # Store the use since we might really need to #include it.
-                if namespace and None not in namespace and '::' not in name:
-                    name = '::'.join(namespace) + '::' + name
-                file_uses[name] = file_uses.get(name, 0) | USES_DECLARATION
                 return
 
-            # TODO(nnorwitz): do proper check for ref/pointer/symbol.
             name = file_use_node[1].filename
-            if name in file_uses:
-                file_uses[name] |= USES_DECLARATION
+            file_uses[name] = file_uses.get(name, 0) | USES_DECLARATION
 
         def _add_variable(node, namespace, reference=False):
             if node.reference or node.pointer or reference:
@@ -288,13 +299,13 @@ class WarningHunter(object):
                               namespace,
                               reference)
 
-        def _process_function(function):
+        def _process_function(function, namespace):
             if function.return_type:
                 return_type = function.return_type
-                _add_variable(return_type, function.namespace)
+                _add_variable(return_type, namespace)
 
             for s in function.specializations:
-                _add_variable(s, function.namespace, not function.body)
+                _add_variable(s, namespace, not function.body)
 
             templated_types = function.templated_types or ()
             for p in function.parameters:
@@ -306,28 +317,32 @@ class WarningHunter(object):
                         # better to iterate through the body and determine
                         # actual uses based on local vars and data members
                         # used.
-                        _add_use(p.type.name, function.namespace)
+                        _add_use(p.type.name, namespace)
                     else:
-                        _add_variable(p.type, function.namespace)
+                        _add_variable(p.type, namespace)
 
         def _process_function_body(function, namespace):
-            iterator = iter(function.body)
-            for t in iterator:
+            previous = None
+            save = namespace[:]
+            for t in function.body:
                 if t.token_type == tokenize.NAME:
+                    previous = t
                     if not keywords.is_keyword(t.name):
-                        # TODO(nnorwitz): handle :: names.
                         # TODO(nnorwitz): handle static function calls.
                         # TODO(nnorwitz): handle using statements in file.
                         # TODO(nnorwitz): handle using statements in function.
                         # TODO(nnorwitz): handle namespace assignment in file.
                         _add_use(t.name, namespace)
-                elif t.name in ('.', '->'):
-                    # Skip tokens after a dereference.
-                    next(iterator)
+                elif t.name == '::' and previous is not None:
+                    namespace.append(previous.name)
+                elif t.name in (':', ';'):
+                    namespace = save[:]
 
         def _add_template_use(name, types, namespace, reference=False):
             for cls in types or ():
-                if name.endswith('_ptr') or cls.pointer or reference:
+                if cls.pointer or cls.reference or reference:
+                    _add_reference(cls.name, namespace)
+                elif name.endswith('_ptr'):
                     # Special case templated classes that end w/_ptr.
                     # These are things like auto_ptr which do
                     # not require the class definition, only decl.
@@ -344,45 +359,41 @@ class WarningHunter(object):
 
         # Iterate through the source AST/tokens, marking each symbols use.
         ast_seq = [self.ast_list]
+        namespace_stack = []
         while ast_seq:
             for node in ast_seq.pop():
                 if isinstance(node, ast.VariableDeclaration):
-                    _add_variable(node.type, node.namespace)
+                    namespace = namespace_stack + node.namespace
+                    _add_variable(node.type, namespace)
                 elif isinstance(node, ast.Function):
-                    _process_function(node)
+                    namespace = namespace_stack + node.namespace
+                    _process_function(node, namespace)
                     if node.body:
-                        _process_function_body(node, node.namespace)
+                        _process_function_body(node, namespace)
                 elif isinstance(node, ast.Typedef):
-                    _process_types(node.alias, node.namespace)
+                    namespace = namespace_stack + node.namespace
+                    _process_types(node.alias, namespace)
                 elif isinstance(node, ast.Friend):
                     expr = node.expr
+                    namespace = namespace_stack + node.namespace
                     if isinstance(expr, ast.Type):
-                        _add_reference(expr.name, node.namespace)
+                        _add_reference(expr.name, namespace)
                     elif isinstance(expr, ast.Function):
-                        _process_function(expr)
+                        _process_function(expr, namespace)
+                elif isinstance(node, ast.Union) and node.body is not None:
+                    ast_seq.append(node.body)
                 elif isinstance(node, ast.Class) and node.body is not None:
-                    if node.body:
-                        ast_seq.append(node.body)
+                    _add_declaration(node.name, node.namespace)
                     _add_template_use('', node.bases, node.namespace)
-                elif isinstance(node, ast.Union) and node.fields:
-                    pass  # TODO(nnorwitz): impl
+                    ast_seq.append(node.body)
+                elif isinstance(node, ast.Using):
+                    if node.names[0].name == 'namespace':
+                        namespace_stack.append(node.names[1].name)
 
         return file_uses, decl_uses
 
     def _find_unused_warnings(self, included_files, forward_declarations,
                               primary_header=None):
-        for node in forward_declarations.values():
-            try:
-                file_use_node = self.symbol_table.lookup_symbol(node.name,
-                                                                node.namespace)
-            except symbols.Error:
-                continue
-            name = file_use_node[1].filename
-            if name in included_files:
-                msg = ("'{}' forward declared, "
-                       "but already #included in '{}'".format(node.name, name))
-                self._add_warning(msg, node)
-
         file_uses, decl_uses = self._determine_uses(included_files,
                                                     forward_declarations)
         if primary_header and primary_header.filename in file_uses:
@@ -390,6 +401,21 @@ class WarningHunter(object):
         self._verify_include_files_used(file_uses, included_files)
         self._verify_forward_declarations_used(forward_declarations, decl_uses,
                                                file_uses)
+        for node in forward_declarations.values():
+            try:
+                file_use_node = self.symbol_table.lookup_symbol(node.name,
+                                                                node.namespace)
+            except symbols.Error:
+                continue
+            name = file_use_node[1].filename
+            if (
+                file_use_node[1].ast_list is not None and
+                name in file_uses and
+                file_uses[name] & USES_DECLARATION
+            ):
+                msg = ("'{}' forward declared, "
+                       "but already #included in '{}'".format(node.name, name))
+                self._add_warning(msg, node)
 
     def _find_header_warnings(self):
         included_files, forward_declarations = self._read_and_parse_includes()
