@@ -27,11 +27,12 @@ __author__ = 'nnorwitz@google.com (Neal Norwitz)'
 
 
 def _find_warnings(filename, lines, ast_list, static_is_optional):
-    def print_warning(node, name):
-        print("{}:{}: static data '{}'".format(
-            filename,
-            lines.get_line_number(node.start),
-            name))
+    def print_warning(node):
+        for name in node.name.split(','):
+            print("{}:{}: static data '{}'".format(
+                filename,
+                lines.get_line_number(node.start),
+                name))
 
     def find_static(function_node):
         tokens = []
@@ -49,28 +50,38 @@ def _find_warnings(filename, lines, ast_list, static_is_optional):
                     tokens = []
                     static_found = False
 
+    count = 0
     for node in ast_list:
         if isinstance(node, ast.VariableDeclaration):
             # Ignore 'static' at module scope so we can find globals too.
             is_static = 'static' in node.type.modifiers
-            is_not_const = 'const' not in node.type.modifiers
+            is_not_const = (
+                'const' not in node.type.modifiers and
+                'constexpr' not in node.type.modifiers
+            )
 
             if is_not_const and (static_is_optional or is_static):
-                print_warning(node, node.name)
+                print_warning(node)
+                count += 1
         elif isinstance(node, ast.Function) and node.body:
             find_static(node)
         elif isinstance(node, ast.Class) and node.body:
             _find_warnings(filename, lines, node.body, False)
 
+    return count
+
+
+def _get_static_declarations(ast_list):
+    for node in ast_list:
+        if (isinstance(node, ast.VariableDeclaration) and
+                'static' in node.type.modifiers):
+            for name in node.name.split(','):
+                yield (name, node)
+
 
 def _find_unused_static_warnings(filename, lines, ast_list):
     """Warn about unused static variables."""
-    static_declarations = {
-        node.name: node
-        for node in ast_list
-        if (isinstance(node, ast.VariableDeclaration) and
-            'static' in node.type.modifiers)
-    }
+    static_declarations = dict(_get_static_declarations(ast_list))
 
     def find_variables_use(body):
         for child in body:
@@ -86,16 +97,23 @@ def _find_unused_static_warnings(filename, lines, ast_list):
                 if isinstance(child, ast.Function) and child.body:
                     find_variables_use(child.body)
 
-    for name in sorted(static_declarations):
+    count = 0
+    for (name, _) in sorted(static_declarations.items(),
+                            key=lambda x: x[1].start):
         if not static_use_counts[name]:
             print("{}:{}: unused variable '{}'".format(
                 filename,
                 lines.get_line_number(static_declarations[name].start),
                 name))
+            count += 1
+
+    return count
 
 
 def run(filename, source, entire_ast, include_paths, quiet):
     lines = metrics.Metrics(source)
 
-    _find_warnings(filename, lines, entire_ast, True)
-    _find_unused_static_warnings(filename, lines, entire_ast)
+    return (
+        _find_warnings(filename, lines, entire_ast, True) +
+        _find_unused_static_warnings(filename, lines, entire_ast)
+    )
